#include <string.h>
#include <values.h>
#include "global.h"
#include "bit_matrix.h"
#include "subset.h"

/* We are using the Set-Covering algorithm from "Set Covering and
 * Involutory bases", by M. Bellmore and H. D. Ratliff, Management
 * Science, vol. 18, pages 194-206, 1971. */

static char is_feasible_without(bit_matrix *b, int *sum, int index)
{
  int j;
  
  for (j=0; j<b->number_of_rows;j++)
    {
      if (!(sum[j] - b->matrix[j][index]))
	return 0;
    }
  
  for (j=0; j<b->number_of_rows;j++)
    sum[j] -= b->matrix[j][index];
  return 1;
}

static char find_nonredundant_feasible_solution(bit_matrix *b, int k, 
						subset *s, subset *Q)
{
  int *sum;
  int i;
  char *bit_array;

  bit_array = (char *)calloc(b->number_of_columns, sizeof(char));
  if (!bit_array)
    return 0;
  
  if (k == 1)
    {
      memset(bit_array, 1, b->number_of_columns*sizeof(char));
    }
  else
    {
      for (i=0; i<s->size; i++)
	bit_array[s->indices[i]] = 1;
      bit_array[Q->indices[0]] = 1;
    }
  subset_zero(s);
  sum = bit_matrix_sum_columns(b, bit_array);
  if (!sum)
    return 0;
  
  for (i=0; i<b->number_of_columns; i++)
    {
      if (bit_array[i])
	bit_array[i] = 1 - is_feasible_without(b, sum, i);
      if (bit_array[i])
	subset_add(s, i);
    }

  free(sum);
  free(bit_array);
  
  return 1;
}

static char find_Q(bit_matrix *b, subset *s, subset *Q, subset *Cb_tag)
{
  int i, j, rows = b->number_of_rows;
  char *check;
  int sum;
  int found = 0;
  
  check = (char *)calloc(s->size, sizeof(char));
  if (!check)
    return 0;
  
  for (i=0; i<rows; i++)
    {
      sum = 0;
      
      for (j=0; j<s->size; j++)
	{
	  if (b->matrix[i][s->indices[j]])
	    {
	      sum++;
	      if (sum > 1)
		break;
	      found = j;
	    }
	  
	}
      
      if ((j==s->size) && !check[found])
	{
	  check[found] = 1;
	  subset_add(Cb_tag, i);
	}
    }

  for (i=0; i<b->number_of_columns; i++)
    {
      sum = 0;
      for (j=0; j<Cb_tag->size; j++)
	{
	  if (b->matrix[Cb_tag->indices[j]][i])
	    sum++;
	  if (sum > 1)
	    break;
	}
      if (sum > 1)
	subset_add(Q, i);
    }

  free(check);
  return 1;
}

int set_covering(relaxation_matrix *r, subset *s_best)
{
  int i, k=1;
  int Z = MAXINT;
  static subset s;
  static char is_first = 1;
  static subset Q;
  static subset Cb_tag;
  int counter = 0;
  
  if (is_first)
    {
      if (!subset_init(&s))
	return 0;
      if (!subset_init(&Q))
	return 0;
      if (!subset_init(&Cb_tag))
	return 0;
      is_first = 0;
    }

  while (1)
    {
      subset_zero(&Q);
      subset_zero(&Cb_tag);

      if (!find_nonredundant_feasible_solution(r->bit, k, &s, &Q))
	return 0;

      if (s.size < Z)
	{
	  Z = s.size;
	  subset_copy(s_best, &s);
	  counter = 0;
	}

      if (!find_Q(r->bit, &s, &Q, &Cb_tag))
	return 0;

      if (Q.size == 0)
	goto Exit;
      
      bit_matrix_add_row(r->bit);
      for (i=0; i<Q.size; i++)
	r->bit->matrix[r->bit->number_of_rows - 1][Q.indices[i]] = 1;
      k++;
      
      counter++;
      if (counter == give_up)
	  return -1;
    }
    
 Exit:
  return 1;
}
